"""Plot statistics using Seaborn."""

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns


def get_bar_x(ax):
    """Get bar centers."""
    x = np.array(sorted([p.get_x() + (p.get_width() / 2) for p in ax.patches]))
    return x


def get_point_max_y(ax):
    """Get highest point in each bin."""
    collections = [c for c in ax.collections if c.get_offsets().shape[0] > 1]
    y = np.array([np.nanmax(c.get_offsets()[:, 1]) for c in collections])
    x = np.array([np.nanmean(c.get_offsets()[:, 0]) for c in collections])
    y = y[np.argsort(x)]
    return y


def add_swarm_bar_sig(ax, sig_ind, y_offset=None):
    """Add significance stats to swarm bar plot."""
    # center of each bar in left/right order
    bx = get_bar_x(ax)

    # highest point in each point plot in left/right order
    cy = get_point_max_y(ax)

    # define y point relative to the highest point
    if y_offset is None:
        y_lim = ax.get_ylim()
        y_offset = (y_lim[1] - y_lim[0]) * .1

    # plot significance markers
    ax.plot(
        bx[sig_ind],
        cy[sig_ind] + y_offset,
        markersize=8,
        color='k',
        linestyle='',
        marker=(5, 2, 0),
    )


def plot_sig(x, y, spacing, line_kws={}, marker_kws={}, ax=None):
    """Plot brackets with a significance indicator."""
    if ax is None:
        ax = plt.gca()

    # bar centers
    x1, x2 = x

    # largest y-values
    y1, y2 = y

    # line and annotation y-values
    yl1 = max(y1, y2) + spacing
    yl2 = yl1 + spacing * .8
    yl3 = yl2 + spacing

    # brackets connecting bars
    line_prop = {'linewidth': 0.75, 'color': 'k'}
    line_prop.update(line_kws)
    ax.plot([x1, x1, x2, x2], [yl1, yl2, yl2, yl1], **line_prop)

    # significance marker
    marker_prop = {'marker': (5, 2, 0), 'markersize': 8, 'color': 'k'}
    marker_prop.update(marker_kws)
    ax.plot((x1 + x2) / 2, yl3, **marker_prop)


def add_swarm_bar_sig_diff(ax, sig_ind_pairs, y_offset=None, **kwargs):
    """Add significant comparison markers to swarm bar plot."""
    bx = get_bar_x(ax)
    cy = get_point_max_y(ax)

    if y_offset is None:
        y_lim = ax.get_ylim()
        y_offset = (y_lim[1] - y_lim[0]) * .05

    for ind in sig_ind_pairs:
        plot_sig(bx[ind], cy[ind], y_offset, ax=ax, **kwargs)


def plot_swarm_bar(
    data=None,
    x=None,
    y=None,
    hue=None,
    dark=None,
    light=None,
    dodge=False,
    point_kind='swarm',
    legend=True,
    width=None,
    sig_ind=None,
    sig_diff_ind=None,
    ax=None,
    point_kws={},
    bar_kws={},
    **kwargs,
):
    """
    Make a bar plot with individual points and error bars.

    Parameters
    ----------
    data : pandas.DataFrame
        Long data frame with variables indicating at least x and y values.

    x : str
        Column in data to determine bar x-positions.

    y : str
        Column in data to determine values to plot.

    hue : str, optional
        Column in data to determine bar and point hues.

    dark : str, optional
        Specification for Seaborn palette; used to plot points.

    light : str, optional
        Specification for Seaborn palette; used to plot bars.

    dodge : bool, optional
        If true, hues will be plotted at different x-values. Used to
        create grouped bar plots.

    point_kind : {'swarm', 'strip'}, optional
        Method for plotting the individual data points in each bin.

    legend : bool, optional
        If true, and the hue input is set, include a legend on the plot.

    width : float, optional
        Width of the bars.

    sig_ind : list of int, optional
        Indices of bars (from left to right) to label as significant.

    sig_diff_ind : list of list of int, optional
        A list of pairs (given as indices of bars from left to right)
        to mark as significant comparisons.

    ax : matplotlib.axes.Axes, optional
        Axes object to draw the plot onto; if not specified, will use
        the current Axes.

    point_kws : dict, optional
        Options when plotting points using seaborn.swarmplot or
        seaborn.stripplot.

    bar_kws : dict, optional
        Options when plotting bars using seaborn.barplot.

    Returns
    -------
    ax : matplotlib.axes.Axes
        Axes that the plot was drawn onto.
    """
    if dark is None:
        dark = 'ch:rot=-.5, light=.7, dark=.3, gamma=.6'
    if light is None:
        light = 'ch:rot=-.5, light=.7, dark=.3, gamma=.2'
    data = data.reset_index()

    if ax is None:
        ax = plt.gca()

    # plot individual points
    point_prop = {
        'size': 3,
        'linewidth': 0.1,
        'edgecolor': 'k',
        'alpha': 0.7,
        'zorder': 3,
        'dodge': dodge,
        'palette': dark,
        'clip_on': False,
    }
    point_prop.update(point_kws)
    if point_kind == 'swarm':
        sns.swarmplot(data=data, x=x, y=y, hue=hue, ax=ax, **point_prop, **kwargs)
    elif point_kind == 'strip':
        sns.stripplot(data=data, x=x, y=y, hue=hue, ax=ax, **point_prop, **kwargs)
    else:
        raise ValueError(f'Invalid point plot kind: {point_kind}')

    # plot error bars for the mean
    bar_prop = {
        'color': 'k',
        'errwidth': 0.8,
        'capsize': 0.425,
        'edgecolor': 'k',
        'linewidth': 0.75,
        'errcolor': 'k',
        'dodge': dodge,
        'palette': light,
    }
    bar_prop.update(bar_kws)
    sns.barplot(data=data, x=x, y=y, hue=hue, ax=ax, **bar_prop, **kwargs)

    # remove overall xlabel and increase size of x-tick labels
    ax.set_xlabel('')
    ax.tick_params(axis='x', labelsize='large')

    # fix ordering of plot elements
    plt.setp(ax.lines, zorder=100, linewidth=1.25, label=None, clip_on=False)
    plt.setp(ax.collections, zorder=100, label=None)

    # modify the width of the bar plot
    if width is not None:
        bar_x = []
        for patch in ax.patches:
            # set width
            center = patch.get_x() + patch.get_width() / 2
            patch.set_width(width)

            # adjust the x position
            patch_x = center - width / 2
            patch.set_x(patch_x)
            bar_x.append(patch_x)
        # set the border around the bars to be the same as the bar spacing
        spacing = 1 - width
        xmin = np.min(bar_x) - spacing
        xmax = np.max(bar_x) + width + spacing
        ax.set_xlim(xmin, xmax)

    # delete legend (redundant with the x-tick labels)
    lg = ax.get_legend()
    if lg is not None:
        lg.remove()
        if hue is not None and legend:
            # refresh the legend to remove the swarm points
            ax.legend()

    # add significance markers
    if sig_ind is not None:
        add_swarm_bar_sig(ax, sig_ind)
    if sig_diff_ind is not None:
        add_swarm_bar_sig_diff(ax, sig_diff_ind)
    return ax



#!/usr/bin/env python
#
# Remove runs that are shorter than the others.

import os
import argparse
import tempfile
import subprocess as sub
import numpy as np
import nibabel as nib
from bids import BIDSLayout


def rename_bold(all_files, included_files, dry_run=False):
    """Delete excluded files and rename included files."""
    example = included_files[0]
    parent_dir = os.path.dirname(example)
    subject = example.tags['subject'].value
    if 'session' in example.tags:
        session = example.tags['session'].value
    else:
        session = None

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_files = []
        n = 0
        for file in all_files:
            if file not in included_files:
                # delete
                cmd = f'rm {file.path}'
                if dry_run:
                    print(cmd)
                else:
                    sub.run(cmd, shell=True)
            else:
                # copy to temp file
                temp = os.path.join(temp_dir, f'file{n}.nii.gz')
                cmd = f'cp {file.path} {temp}'
                if dry_run:
                    print(cmd)
                else:
                    sub.run(cmd, shell=True)
                n += 1
                temp_files.append(temp)

        for i, (file, temp) in enumerate(zip(included_files, temp_files)):
            # set new file name
            run = i + 1
            task = file.tags['task'].value
            if session is not None:
                prefix = f'sub-{subject}_ses-{session}'
            else:
                prefix = f'sub-{subject}'
            filename = f'{prefix}_task-{task}_run-{run}_bold.nii.gz'

            # copy from the temp file to the new path
            new_file = os.path.join(parent_dir, filename)
            cmd = f'cp {temp} {new_file}'
            if dry_run:
                print(cmd)
            else:
                sub.run(cmd, shell=True)


def main(data_dir, dry_run=False):
    layout = BIDSLayout(data_dir)
    subjects = layout.get_subjects()
    for subject in subjects:
        bold_files = layout.get(
            subject=subject, datatype='func', extension='.nii.gz', suffix='bold'
        )
        n = np.zeros(len(bold_files))
        for i, bold_file in enumerate(bold_files):
            img = nib.load(bold_file)
            n[i] = img.header.get_data_shape()[3]

        max_n = np.max(n)
        if np.all(n == max_n):
            continue

        include = n == max_n
        included_files = [f for f, i in zip(bold_files, include) if i]
        rename_bold(bold_files, included_files, dry_run)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('data_dir', help="Path to raw BIDS data")
    parser.add_argument(
        '--dry-run', action="store_true", help="Print commands without running"
    )
    args = parser.parse_args()
    main(args.data_dir, args.dry_run)



"""Display task searchlight analysis in brainiak."""

import warnings
import numpy as np
import pandas as pd
import scipy.stats as stats
import scipy.optimize as optim
from scipy.spatial.distance import pdist, cdist, squareform


def labels2rdm(labels):
    """Create an RDM from a vector with categorical labels."""
    n = len(labels[0])
    # get pairs where all labels are equal
    conj = np.ones((n, n), dtype=bool)
    for label in labels:
        iseq = label[:, np.newaxis] == label.flatten()
        conj = np.logical_and(conj, iseq)

    # make into an RDM
    mat = 1 - np.asarray(conj, "float")

    mat[np.diag_indices(mat.shape[0])] = 0
    return mat


def pair_and(x, y=None):
    """Pairs where conditions are both true."""
    if y is None:
        y = x
    return np.asarray(x)[:, None] & np.asarray(y)[:, None].T


def pair_eq(x, y=None):
    """Pairs where conditions are equal."""
    if y is None:
        y = x
    return np.asarray(x)[:, None] == np.asarray(y)[:, None].T


def pair_neq(x, y=None):
    """Pairs where conditions are not equal."""
    if y is None:
        y = x
    return np.asarray(x)[:, None] != np.asarray(y)[:, None].T


def perm_z(stat_perm):
    """Calculate a z-statistic from permutation results."""
    if np.any(np.isnan(stat_perm) | np.isinf(stat_perm)):
        warnings.warn("Invalid permutation statistic.", RuntimeWarning)
        return np.nan
    elif len(np.unique(stat_perm)) == 1:
        warnings.warn("Permutation statistic does not vary.", RuntimeWarning)
        return np.nan

    p = np.mean(stat_perm >= stat_perm[0])
    e = 1.0 / len(stat_perm)
    if p > (1 - e):
        p = 1 - e
    return stats.norm.ppf(1 - p)


def init_pRSA(n_perm, model_rdms, rank=True):
    """Prepare model RDMs for partial RSA.

    Parameters
    ----------
    n_perm : int
        Number of model permutations to test to establish a null
        distribution for each partial correlation.

    model_rdms : list of numpy arrays
        Representational dissimilarity matrix for each of a set of
        models to test. Each matrix should be [trials x trials],
        giving the dissimilarity between trials predicted by that model.

    rank : bool
        If true, RDMs will be rank transformed.

    Returns
    -------
    rsa_def : dict
        Information needed to run the partial RSA. Includes:
        rsa_def['model_mats'] : list of numpy arrays
            Design matrix for each model. This includes rank-transformed
            representational dissimilarity vectors for all of the other
            models.
        rsa_def['model_resid'] : list of numpy arrays
            Residual vectors for each model after rank transforming and
            controlling for all the other models.
    """

    n_model = len(model_rdms)
    n_item = model_rdms[0].shape[0]

    # generate random indices to use across all models
    rand_ind = [
        np.random.choice(np.arange(n_item), n_item, False) for i in range(n_perm)
    ]
    rand_ind.insert(0, np.arange(n_item))

    # put models in vector format and rank-order them
    if rank:
        model_vecs = np.asarray(
            [stats.rankdata(squareform(rdm)) for rdm in model_rdms]
        ).T
    else:
        model_vecs = np.asarray([squareform(rdm) for rdm in model_rdms]).T

    # make control design matrices for regression
    model_mats = []
    for i in range(n_model):
        other = [j for j in range(n_model) if j is not i]
        control_vecs = model_vecs[:, other]
        intercept = np.ones((control_vecs.shape[0], 1))
        mat = np.hstack((control_vecs, intercept))
        model_mats.append(mat)

    # residualize each model against the others
    model_resid = []
    for i in range(n_model):
        resid = []
        for ind in rand_ind:
            rand_rdm = model_rdms[i][np.ix_(ind, ind)]
            rand_model = squareform(rand_rdm)
            if rank:
                rand_model = stats.rankdata(rand_model)
            beta = optim.nnls(model_mats[i], rand_model)[0]
            res = rand_model - model_mats[i].dot(beta)
            resid.append(res)
        model_resid.append(np.asarray(resid))

    return {"model_mats": model_mats, "model_resid": model_resid}


def perm_partial(data_vec, model_mat, model_resid):
    """Calculate partial correlation for multiple permutations."""

    # regress to get residuals from data similarity vector
    beta = optim.nnls(model_mat, data_vec)[0]
    data_resid = data_vec - model_mat.dot(beta)

    # correlate with the residualized (random) model
    xmat = data_resid.reshape((1, len(data_resid)))
    stat = 1 - cdist(xmat, model_resid, "correlation").squeeze()
    if np.any(np.isnan(stat)):
        raise ValueError("statistic is undefined.")
    return stat


def call_pRSA(subj, mask, sl_rad, bcast_var, rank=True):
    """Function to pass when running partial RSA searchlight.

    Parameters
    ----------
    subj : list of numpy arrays
        Each element contains a 4D array of functional data for one
        subject.

    mask : numpy array
        A 3D array with the current mask for the searchlight.

    sl_rad : int
        Searchlight radius.

    bcast_var : dict
        The dict created by init_pRSA.

    rank : bool
        If true, rank transform data.

    Returns
    -------
    stat_all : tuple
        The z-statistic for each of the partial correlations, compared
        to the null permutation distribution.
    """

    # data representational dissimilarity vector
    data = subj[0][mask, :].T
    data_vec = pdist(data, "correlation")
    if rank:
        data_vec = stats.rankdata(data_vec)

    # unpack global data
    model_mats = bcast_var["model_mats"]
    model_resid = bcast_var["model_resid"]

    # calculate z-statistic for each partial correlation
    stat_all = []
    for i in range(len(model_mats)):
        stat = perm_partial(data_vec, model_mats[i], model_resid[i])
        stat_all.append(perm_z(stat))

    return tuple(stat_all)


def sign_perm(mat, n_perm, beta):
    """Test for significantly positive values with FDR correction."""
    # generate random sign flips
    n_samp, n_test = mat.shape
    rand_sign = np.hstack(
        (np.ones((n_samp, 1)), np.random.choice([-1, 1], (n_samp, n_perm)))
    )

    # apply random sign to all conditions
    mat_perm = mat[:, :, None] * rand_sign[:, None, :]

    # ttest vs zero for permuted and actual stats
    perm_t, perm_p = stats.ttest_1samp(mat_perm, 0, axis=0)
    t, p = stats.ttest_1samp(mat_perm[:, :, 0], 0, axis=0)

    # convert to right-tailed test
    perm_p /= 2
    perm_p[perm_t < 0] = 1 - perm_p[perm_t < 0]
    p /= 2
    p[t < 0] = 1 - p[t < 0]

    # q-value at each actual p-value
    ERstar = np.zeros(p.shape)
    r = np.zeros(p.shape)
    rstarb = np.zeros(p.shape)
    m = n_test
    q = np.zeros(p.shape)
    for i, a in enumerate(p):
        # expected value of number of rejections under the null
        Rstar = np.sum(perm_p <= a, 0)
        ERstar[i] = np.mean(Rstar)

        # 1 - beta quantile of Rstar
        rstarb[i] = np.quantile(Rstar, 1 - beta)

        # number of actual rejections
        r[i] = np.sum(p <= a)

        # FDR local estimator
        if (r[i] - rstarb[i]) >= (a * m):
            q[i] = ERstar[i] / (ERstar[i] + r[i] - (a * m))
        else:
            q[i] = np.mean(Rstar >= 1)

    # package results
    std = np.std(mat, axis=0, ddof=1)
    mean = np.mean(mat, axis=0)
    sem = stats.sem(mat, axis=0)
    df = pd.DataFrame({
        't': t, 'p': p, 'q': q, 'ERstar': ERstar, 'rstarb': rstarb, 'r': r,
        'mean': mean, 'std': std, 'sem': sem, 'd': np.abs(mean) / std
    })
    return df



"""Utilities for plotting dissimilarity matrices."""

import math
import numpy as np
from scipy import stats
from scipy import spatial
import scipy.spatial.distance as sd
from sklearn import manifold
import matplotlib.pyplot as plt
from matplotlib.offsetbox import OffsetImage, AnnotationBbox


def plot_pattern(mat, ax=None):
    """Plot a matrix as a simple grayscale image with no axes."""
    if ax is None:
        ax = plt.gca()
    h = ax.matshow(mat, cmap="gray")
    ax.set_axis_off()
    return h


def plot_dsm(
        dsm, rank=False, prange=(1, 99), vlim=None, cmap="viridis", checks=True,
        ax=None
):
    """Plot pairwise dissimilarity values as a matrix."""
    if ax is None:
        ax = plt.gca()

    if rank:
        dsm = sd.squareform(stats.rankdata(sd.squareform(dsm, checks=checks)))

    if vlim is None:
        vlim = np.percentile(sd.squareform(dsm, checks=checks), prange)

    h = ax.pcolor(dsm, vmin=vlim[0], vmax=vlim[1], cmap=cmap)
    ax.set_aspect(1)
    if not ax.yaxis_inverted():
        ax.invert_yaxis()
    ax.set_axis_off()
    return h


def embed(dsm, embedding=None):
    """Calculate a two-dimensional MDS embedding for a dissimilarity matrix."""
    if embedding is None:
        embedding = manifold.MDS(n_components=2, dissimilarity="precomputed")
    X = embedding.fit_transform(dsm)
    x, y = X.T
    return x, y


def embed_align(dsm_set, reference=0, embedding=None):
    """Calculate embedding for dissmilarity matrices and align them."""
    embed_set = [np.vstack(embed(dsm, embedding)).T for dsm in dsm_set]
    ref_embed = embed_set[reference]
    aligned = [None] * len(dsm_set)
    for i, mat in enumerate(embed_set):
        if i == reference:
            continue
        pro1, pro2, disparity = spatial.procrustes(ref_embed, mat)
        if aligned[reference] is None:
            aligned[reference] = pro1
        aligned[i] = pro2
    return aligned


def pad_image(image, shape, value):
    """Pad an image to match a given shape."""
    # center align to fill difference in shape
    full_shape = np.hstack((shape, image.shape[-1]))
    width = full_shape - image.shape
    pad_width = [(math.floor(w / 2), math.ceil(w / 2)) for w in width]
    padded = np.pad(image, pad_width, constant_values=value)
    return padded


def image_matrix(images, shape, background=1):
    """Concatenate stimulus images into a matrix."""
    # get the maximum image size
    image_shapes = np.array([i.shape for i in images])
    max_size = np.max(image_shapes, 0)

    # compile the stimulus matrix
    ind = 0
    mat_rows = []
    for i in range(shape[0]):
        mat_cols = []
        for j in range(shape[1]):
            # pad the image as needed to make everything constant size
            image = images[ind]
            padded = pad_image(image, max_size[:2], background)
            mat_cols.append(padded)
            ind += 1
        mat_rows.append(np.hstack(mat_cols))
    return np.vstack(mat_rows)


def pool_image_matrix(pool, images, shape, **kwargs):
    """Plot an image matrix for a pool."""
    items = pool["stim"].to_list()
    pool_images = [images[item] for item in items]
    mat = image_matrix(pool_images, shape, **kwargs)
    return mat


def image_scatter(x, y, images, ax=None, zoom=1.0, frameon=False):
    """Make a scatter plot with images instead of points."""
    if ax is None:
        ax = plt.gca()

    artists = []
    x, y = np.atleast_1d(x, y)
    for (x0, y0, im) in zip(x, y, images):
        im0 = OffsetImage(im, zoom=zoom)
        ab = AnnotationBbox(im0, (x0, y0), xycoords="data", frameon=frameon, pad=0)
        artists.append(ax.add_artist(ab))
    ax.update_datalim(np.column_stack([x, y]))
    ax.autoscale()
    return artists


def plot_mds(dsm, images, ind=None, embedding=None, zoom=1.0, ax=None):
    """Plot multidimensional scaling of a dissimilarity matrix."""
    if ind is not None:
        if embedding == "precomputed":
            dsm = dsm[:, ind]
        else:
            dsm = dsm[ind, ind]
        images = images[ind]

    if embedding == "precomputed":
        x, y = dsm
    else:
        x, y = embed(dsm, embedding)

    if np.ptp(x) >= np.ptp(y):
        d1, d2 = x, y
    else:
        d1, d2 = y, x
    artists = image_scatter(d1, d2, images, ax=ax, zoom=zoom)
    return artists


def plot_mds_aligned(
    model_set, model_names, images, embedding=None, fig=None, col_wrap=3
):
    """Plot MDS aligned over multiple models."""
    mset = [model_set[name] for name in model_names]
    aligned = embed_align(mset, embedding)

    if fig is None:
        fig = plt.figure(figsize=(14, 7))

    n_row = int(np.ceil(len(aligned) / col_wrap))
    for i, mat in enumerate(aligned):
        ax = fig.add_subplot(n_row, col_wrap, i + 1)
        plot_mds(mat.T, images, zoom=0.06, ax=ax, embedding="precomputed")
        ax.set_axis_off()
        ax.set_aspect("equal")
        ax.set_title(model_names[i])
    return fig


def plot_models_dsm(model_dsm, fig=None, prange=(1, 99), col_wrap=None):
    """Plot multiple dissimilarity matrices."""
    if fig is None:
        fig = plt.gcf()

    n_model = len(model_dsm)
    if col_wrap is None:
        n_row = 1
        n_col = n_model
    else:
        n_col = col_wrap
        n_row = math.ceil(n_model / col_wrap)

    i = 0
    for name, dsm in model_dsm.items():
        ax = fig.add_subplot(n_row, n_col, i + 1)
        plot_dsm(dsm, ax=ax, prange=prange)
        ax.set_title(name)
        i += 1


def plot_rep_as_dsm(
    data=None, distance="correlation", ax=None, prange=(1, 99), color=None, **plot_kws
):
    """Convert representation to dissimilarity and plot."""
    mat = data.filter(like="dim").to_numpy()
    dsm = sd.squareform(sd.pdist(mat, distance))
    plot_dsm(dsm, ax=ax, prange=prange, **plot_kws)


def plot_rep_as_mds(
    data=None, distance="correlation", ax=None, images=None, color=None, zoom=1,
    embedding=None
):
    """Run dimensionality reduction and plot images."""
    mat = data.filter(like="dim").to_numpy()
    dsm = sd.squareform(sd.pdist(mat, distance))
    x, y = embed(dsm, embedding)
    im_list = [images[stim] for stim in data.stim.to_list()]
    image_scatter(x, y, im_list, ax=ax, zoom=zoom)



"""Utilities for working with subject data."""

import os
import re
import time
import subprocess as sub
from datetime import datetime
from glob import glob
from argparse import ArgumentParser


def imname(filepath):
    """Get the name of an .nii.gz file."""
    return os.path.basename(filepath.split(".nii.gz")[0])


def impath(*args):
    """Construct the path to an .nii.gz file."""
    p = os.path.join(*args)
    if not p.endswith(".nii.gz"):
        p += ".nii.gz"
    return p


class SubjParser(ArgumentParser):
    """Class for parsing standard arguments."""

    def __init__(self, include_log=True, raw=False, **init_args):
        if raw:
            from argparse import RawTextHelpFormatter

            ArgumentParser.__init__(
                self, formatter_class=RawTextHelpFormatter, **init_args
            )
        else:
            ArgumentParser.__init__(self, **init_args)
        self.add_argument(
            "subject", type=str, help="name of subject directory in study directory"
        )
        if "STUDYDIR" in os.environ:
            study_dir = os.environ["STUDYDIR"]
        else:
            study_dir = None

        s = """path to main study directory; if not set, value of STUDYDIR
environment variable will be used"""
        self.add_argument("--study-dir", type=str, default=study_dir, help=s)
        if include_log:
            self.add_argument(
                "--dry-run",
                help="display commands without executing",
                default=False,
                action="store_true",
            )
            self.add_argument(
                "--clean-logs",
                help="remove existing similar logs",
                default=False,
                action="store_true",
            )


class SubjLog:
    """Class for logging subject processing."""

    def __init__(
        self, subject, base, main=None, rm_existing=False, dry_run=False, study_dir=None
    ):

        self.subject = subject
        self.name = base
        self.dry_run = dry_run
        self.main_file = None
        self.start_time = None

        # set the log file
        if study_dir is None:
            study_dir = os.environ["STUDYDIR"]
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        filename = base + "_" + timestamp + ".log"
        log_dir = os.path.join(study_dir, subject, "logs")
        if not os.path.exists(log_dir):
            os.mkdir(log_dir)
        log_file = os.path.join(log_dir, filename)
        if main is not None:
            filename = main + ".log"
            self.main_file = os.path.join(log_dir, filename)

        if rm_existing:
            # clear logs that match the supplied base
            existing = glob(os.path.join(log_dir, "%s_*.log" % base))
            for filepath in existing:
                os.remove(filepath)
        self.log_file = log_file

    def timestamp(self):
        """Get a timestamp with standard formatting for a log."""
        return datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    def start(self):
        """Start logging."""

        if self.dry_run:
            return

        # print a header
        msg = "Starting %s for %s at: %s\n" % (
            self.name,
            self.subject,
            self.timestamp(),
        )
        self.start_time = time.time()
        self.write(msg, wrap=False)
        if self.main_file:
            self.write(msg, wrap=False, main_log=True)

    def finish(self):
        """Finish and close the log."""

        if self.dry_run:
            return

        finish = time.time() - self.start_time
        msg = "Finished %s for %s at: %s\n" % (
            self.name,
            self.subject,
            self.timestamp(),
        )
        self.write("\n" + msg, wrap=False)
        self.write("Took %d s." % finish, wrap=False)
        if self.main_file:
            self.write(msg, wrap=False, main_log=True)

    def run(self, cmd):
        """Run a command with input and output logging."""

        print(cmd)
        if self.dry_run:
            return

        # open log file and print command to run
        outfile = open(self.log_file, "a")
        outfile.write("\n" + cmd + "\n")
        outfile.close()

        # actually running the command
        p = sub.Popen(cmd, stdout=sub.PIPE, stderr=sub.PIPE, shell=True)
        output, errors = p.communicate()
        outfile = open(self.log_file, "a")
        if output:
            print(output)
            outfile.write(output)
        if errors:
            outfile.write("ERROR: " + errors)
            print("%s: ERROR: " % self.subject + errors)
        outfile.close()

    def write(self, message, wrap=True, main_log=False):
        """Write a message to the log."""

        if self.dry_run:
            print(message)
            return

        if main_log:
            outfile = open(self.main_file, "a")
        else:
            outfile = open(self.log_file, "a")
        if wrap:
            outfile.write("\nMESSAGE: " + message + "\n")
        else:
            outfile.write(message)
        outfile.close()


class SubjPath:
    """Information about subject directory structure."""

    def __init__(self, subject, study_dir=None):

        self.subject = subject
        if study_dir is not None:
            self.study_dir = study_dir
        else:
            self.study_dir = os.environ["STUDYDIR"]
        self.subj_dir = os.path.join(self.study_dir, self.subject)
        self.dirnames = [
            "anatomy",
            "behav",
            "BOLD",
            "DTI",
            "fieldmap",
            "logs",
            "model",
            "raw",
        ]
        self.d = dict()
        self.d["base"] = self.subj_dir
        for dirname in self.dirnames:
            self.d[dirname.lower()] = os.path.join(self.subj_dir, dirname)

    def __str__(self):
        return "Subject %s (%s)" % (self.subject, self.subj_dir)

    def make_std_dirs(self):
        """Make the set of standard subject directories."""

        if not os.path.exists(self.d["base"]):
            os.mkdir(self.d["base"])
        for std in self.dirnames:
            name = std.lower()
            if not os.path.exists(self.d[name]):
                os.mkdir(self.d[name])

    def path(self, std, *args):
        """Get the path to file or directory within a standard directory."""
        if std in self.d:
            fulldir = os.path.join(self.d[std.lower()], *args)
        else:
            fulldir = os.path.join(self.subj_dir, std, *args)
        return fulldir

    def image_path(self, std, *args):
        """Get the path to an image file for a subject."""
        return impath(self.path(std, *args))

    def proj_path(self, *args):
        """Path to a file within the code project."""
        proj_dir = os.path.dirname(os.path.dirname(__file__))
        fulldir = os.path.join(proj_dir, *args)
        return fulldir

    def glob(self, std, *args):
        """Get all files or directories matching a pattern with *."""
        paths = glob(os.path.join(self.d[std.lower()], *args))
        return paths

    def match_dirs(self, main_dir, dir_pattern):
        """Get directories matching a regular expression."""
        dirs = os.listdir(main_dir)
        test = re.compile(dir_pattern)
        match = []
        for d in dirs:
            full = os.path.join(main_dir, d)
            if test.match(d) and os.path.isdir(full):
                match.append(full)
        match.sort()
        return match

    def init_log(self, base, main, args):
        """Initialize a log for processing this subject."""
        log = SubjLog(
            self.subject,
            base,
            main,
            rm_existing=args.clean_logs,
            dry_run=args.dry_run,
            study_dir=args.study_dir,
        )
        return log



"""General linear modeling of fMRI data."""

from pathlib import Path
import logging
import json
import numpy as np
import pandas as pd
import nibabel as nib
import matplotlib as mpl
import matplotlib.pyplot as plt
from nilearn.glm import first_level
from nipype.interfaces import afni
import click


def create_simple_design(events, trial_column, n_vol, tr, time_offset, high_pass=0):
    """
    Create a simple design matrix.

    Create a linear model design matrix for a "simple" design, i.e.,
    one that only has one event regressor to be modeled.

    Parameters
    ----------
    events : pandas.DataFrame
        BIDS-compliant events table with onset and duration columns.

    trial_column : str
        Column of events to convolve with a hemodynamic response function.

    n_vol : int
        Number of volumes in the timeseries to be modeled.

    tr : float
        Repetition time, i.e., the sampling rate of the timeseries to
        be modeled, in seconds.

    time_offset : float
        Start time of the first timeseries sample, in seconds.

    high_pass : float
        Cutoff frequency for the high-pass filter in Hz.

    Returns
    -------
    design : pandas.DataFrame
        Design matrix with a convolved event regressor and high-pass filter
        nuisance regressors.
    """

    # check for required columns in events
    required_columns = ["onset", "duration", trial_column]
    for column in required_columns:
        if column not in events:
            raise ValueError(f"Must define {column} column in events.")

    # create dataframe with trial information
    trials = pd.DataFrame(
        {
            "onset": events["onset"],
            "duration": events["duration"],
            "trial_type": events[trial_column],
        }
    )

    # create a design matrix
    frame_times = time_offset + np.arange(n_vol) * tr
    design = first_level.make_first_level_design_matrix(
        frame_times, events=trials, high_pass=high_pass
    )
    return design


def create_confound_matrix(
    confounds, regressors=None, censor_motion=False, censor_motion_range=(-1, 2)
):
    """
    Prepare confound matrix from fMRIPrep output.

    Parameters
    ----------
    confounds : pandas.DataFrame
        Table of confound timeseries output by fMRIPrep.

    regressors : list of str
        Confound regressors to include. Each regressor will be
        mean-centered.

    censor_motion : bool
        If true, high-motion volumes will be censored (i.e., a
        regressor will be added for each censored volumne to regress
        out any activation specific to that timepoint).

    censor_motion_range : tuple of int
        First and last volumes to censor, relative to high-motion
        volumes at time zero (e.g., -1 indicates volume before high-
        motion volume, 2 indicates two timepoints after).

    Returns
    -------
    nuisance : numpy.ndarray
        Matrix of nuisance regressor timeseries.

    nuisance_names : list of str
        Name of each included nuisance regressor.
    """
    nuisance_names = []

    # create nuisance regressor matrix
    if regressors is not None:
        raw = confounds.get(regressors).to_numpy()
        nuisance = raw - np.nanmean(raw, 0)
        nuisance[np.isnan(nuisance)] = 0
        nuisance_names.extend(regressors)
    else:
        nuisance = None

    # exclude motion outliers
    if censor_motion:
        outliers = confounds.filter(like="motion_outlier")
        if not outliers.empty:
            # get high-motion times
            motion_ind = np.where(outliers)[0]

            # add offsets to create window around those times
            offsets = np.arange(censor_motion_range[0], censor_motion_range[1] + 1)
            censor_ind = np.unique(motion_ind[:, np.newaxis] + offsets)
            reg_ind = np.arange(len(censor_ind))

            # assign separate regressor for each censored time point
            mat = np.zeros((outliers.shape[0], len(censor_ind)))
            mat[censor_ind, reg_ind] = 1
            if nuisance is not None:
                nuisance = np.hstack([nuisance, mat])
            else:
                nuisance = mat
            censor_names = [f"motion_{i + 1:03}" for i in range(mat.shape[1])]
            nuisance_names.extend(censor_names)
    return nuisance, nuisance_names


def estimate_betaseries(data, design, confound=None, category=None):
    """
    Estimate beta images for a set of trials.

    Parameters
    ----------
    data : numpy.ndarray
        [timepoints x voxels] array of functional data to model.

    design : numpy.ndarray
        [timepoints x EVs] array. Each explanatory variable (EV) will
        be estimated using a separate model.

    confound : numpy.ndarray
        [timepoints x regressors] array with regressors of no interest,
        which will be included in each model.

    category : numpy.ndarray
        [EVs] array with the category label of each EV.

    Returns
    -------
    beta : numpy.ndarray
        [EVs by voxels] array of beta estimates.
    """
    n_trial = design.shape[1]
    n_sample = data.shape[0]
    beta_maker = np.zeros((n_trial, n_sample))
    trial_evs = list(range(n_trial))
    for i, ev in enumerate(trial_evs):
        # this trial
        dm_trial = design[:, ev, np.newaxis]

        # other trials, summed together
        if category is not None:
            ucat = np.unique(category)
            dm_otherevs = np.zeros((design.shape[0], len(ucat)))
            for j, cat in enumerate(ucat):
                other_evs = [
                    x for x, c in zip(trial_evs, category) if x != ev and c == cat
                ]
                dm_otherevs[:, j] = np.sum(design[:, other_evs], 1)
        else:
            other_trial_evs = [x for x in trial_evs if x != ev]
            dm_otherevs = np.sum(design[:, other_trial_evs, np.newaxis], 1)

        # put together the design matrix
        if confound is not None:
            dm_full = np.hstack((dm_trial, dm_otherevs, confound))
        else:
            dm_full = np.hstack((dm_trial, dm_otherevs))
        s = dm_full.shape
        dm_full = dm_full - np.kron(np.ones(s), np.mean(dm_full, 0))[: s[0], : s[1]]
        dm_full = np.hstack((dm_full, np.ones((n_sample, 1))))

        # calculate beta-forming vector
        beta_maker_loop = np.linalg.pinv(dm_full)
        beta_maker[i, :] = beta_maker_loop[0, :]

    # this uses Jeanette Mumford's trick of extracting the beta-forming
    # vector for each trial and putting them together, which allows
    # estimation for all trials at once
    beta = np.dot(beta_maker, data)
    return beta


def run_betaseries(
    events_path,
    events_field,
    tr,
    time_offset,
    func_path,
    mask_name,
    mask_file,
    out_dir,
    subject,
    task,
    run,
    space,
    events_category=None,
    nuisance=None,
    nuisance_names=None,
    high_pass=0,
    smooth_fwhm=None,
    sort_field=None,
):
    if sort_field is None:
        sort_field = events_field

    # load events
    events = pd.read_table(events_path, dtype={"group": "str"})

    # add an EV name column (necessary to avoid name restrictions in nilearn)
    evs = events.sort_values(sort_field)[events_field].unique()
    n_ev = len(evs)
    ev_inds = np.arange(n_ev)
    ev_names = [f"ev_{e + 1:03}" for e in ev_inds]
    events["ev"] = events[events_field]
    events["ev_index"] = events[events_field].map(dict(zip(evs, ev_names)))

    # create design matrix
    img = nib.load(func_path)
    n_vol = img.header["dim"][4]
    design = create_simple_design(
        events, "ev_index", n_vol, tr, time_offset, high_pass=high_pass
    )
    design_header = design.columns.to_list()[:-1]

    # get corresponding events; keep any fields that are consistent across
    # presentations
    n = events.groupby("ev_index").apply(lambda x: x.nunique()).reset_index(drop=True)
    consistent = (n == 1).all()
    consistent_fields = consistent[consistent].index
    ev_events = (
        events.groupby("ev_index")
        .first()
        .reset_index()
        .get(consistent_fields)
        .drop(columns=["ev", "ev_index"])
    )
    if events_category is not None:
        category = ev_events[events_category]
    else:
        category = None

    # create confound matrix
    mat = design.iloc[:, :n_ev].to_numpy()
    if nuisance is not None:
        confound = np.hstack((design.iloc[:, n_ev:-1].to_numpy(), nuisance))
        design_header.extend(nuisance_names)
    else:
        confound = design.iloc[:, n_ev:-1]
    full_design = pd.DataFrame(np.hstack([mat, confound]), columns=design_header)

    # prepare output directory
    sub_path = Path(out_dir) / f"sub-{subject}" / "func"
    sub_path.mkdir(parents=True, exist_ok=True)
    prefix = f"sub-{subject}_task-{task}_run-{run}"

    # smooth functional data within mask
    if smooth_fwhm is not None:
        smooth_path = (
            sub_path
            / f"{prefix}_space-{space}_label-{mask_name}_desc-smooth_bold.nii.gz"
        )
        logging.getLogger("nipype.interface").setLevel(0)
        bim = afni.BlurInMask()
        bim.inputs.in_file = func_path
        bim.inputs.mask = mask_file
        bim.inputs.fwhm = smooth_fwhm
        bim.inputs.out_file = smooth_path
        bim.inputs.options = "-overwrite"
        bim.run()
        func_path = smooth_path

    # load functional data
    bold_vol = nib.load(func_path)
    mask_vol = nib.load(mask_file)
    bold_img = bold_vol.get_fdata()
    mask_img = mask_vol.get_fdata().astype(bool)
    data = bold_img[mask_img].T

    # estimate betaseries
    beta = estimate_betaseries(data, mat, confound=confound, category=category)
    out_data = np.zeros([*mask_img.shape, beta.shape[0]])
    out_data[mask_img, :] = beta.T

    # save betaseries image
    beta_path = sub_path / f"{prefix}_space-{space}_label-{mask_name}_betaseries.nii.gz"
    new_img = nib.Nifti1Image(out_data, mask_vol.affine, mask_vol.header)
    nib.save(new_img, beta_path)

    # save mask
    mask_path = sub_path / f"{prefix}_space-{space}_label-{mask_name}_mask.nii.gz"
    new_img = nib.Nifti1Image(mask_img, mask_vol.affine, mask_vol.header)
    nib.save(new_img, mask_path)

    # save design
    design_path = sub_path / f"{prefix}_desc-design_timeseries.tsv"
    full_design.to_csv(design_path, sep="\t", index=False, na_rep="n/a")
    mpl.use("Agg")
    fig, ax = plt.subplots()
    ax.axis("off")
    full_mat = full_design.to_numpy()
    full_min = np.min(full_mat, axis=0)
    full_max = np.max(full_mat, axis=0)
    ax.matshow((full_mat - full_min) / (full_max - full_min))
    fig.savefig(design_path.with_suffix(".png"), pad_inches=0, dpi=600)

    # save corresponding events
    evs_path = sub_path / f"{prefix}_desc-events_timeseries.tsv"
    ev_events.to_csv(evs_path, sep="\t", index=False, na_rep="n/a")


@click.command()
@click.argument("bold_file", type=click.Path(exists=True))
@click.argument("tr", type=float)
@click.argument("time_offset", type=float)
@click.argument("mask_name", type=str)
@click.argument("mask_file", type=click.Path(exists=True))
@click.argument("events_file", type=click.Path(exists=True))
@click.argument("events_field", type=str)
@click.argument("out_dir", type=click.Path())
@click.argument("subject", type=str)
@click.argument("task", type=str)
@click.argument("run", type=str)
@click.argument("space", type=str)
@click.option(
    "--events-category",
    type=str,
    help="Field with event category (categories are modeled as separate regressors)",
)
@click.option(
    "--sort-field",
    type=str,
    help="Colon-separated list of fields to sort by for output order",
)
@click.option("--high-pass", help="Highpass filter in Hz", type=float)
@click.option("--smooth", help="Smoothing kernel FWHM", type=float)
@click.option(
    "--confound-file", help="Path to confound matrix file", type=click.Path(exists=True)
)
def betaseries(
    bold_file,
    tr,
    time_offset,
    mask_name,
    mask_file,
    events_file,
    events_field,
    out_dir,
    subject,
    task,
    run,
    space,
    events_category,
    sort_field,
    high_pass,
    smooth,
    confound_file,
):
    """
    Create a betaseries image estimating stimulus activity for a functional run.

    This script is designed to be flexible in where data are located, but assumes
    that confound regressors are pre-made and stored in a text file.

    \b
    Required inputs
    ---------------
    bold_file
        Path to NIfTI file with functional timeseries data.
    tr
        Repetition time in s.
    time_offset
        Time in s that frames were collected, relative to time zero.
    mask_name
        Name of the mask to use when selecting voxels to analyze.
    mask_file
        Path to a binary mask indicating voxels to analyze.
    events_file
        Path to a tab- or comma-separated values file with task events. Must
        include onset and duration fields (see BIDS specification for task events).
    events_field
        Column of the events file to use to indicate individual EVs (explanatory
        variables) to model.
    out_dir
        Path to main output directory.
    subject
        Subject ID.
    task
        Task code.
    run
        Run number.
    space
        Space in which the data should be analyzed (must have been a space output
        by fMRIPrep).
    """
    if sort_field is None:
        sort_field = events_field
    else:
        sort_field = sort_field.split(":")

    # load nuisance regressors
    if confound_file is not None:
        nuisance = np.loadtxt(confound_file)
        nuisance_names = [f"confound_{i + 1:03}" for i in range(nuisance.shape[1])]
    else:
        nuisance = None
        nuisance_names = []

    # run betaseries estimation and save results
    run_betaseries(
        events_file,
        events_field,
        tr,
        time_offset,
        bold_file,
        mask_name,
        mask_file,
        out_dir,
        subject,
        task,
        run,
        space,
        events_category=events_category,
        nuisance=nuisance,
        nuisance_names=nuisance_names,
        high_pass=high_pass,
        smooth_fwhm=smooth,
        sort_field=sort_field,
    )


@click.command()
@click.argument("data_dir", type=click.Path(exists=True))
@click.argument("fmriprep_dir", type=click.Path(exists=True))
@click.argument("out_dir", type=click.Path())
@click.argument("subject", type=str)
@click.argument("task", type=str)
@click.argument("run", type=str)
@click.argument("space", type=str)
@click.argument("mask_name", type=str)
@click.argument("mask_file", type=click.Path(exists=True))
@click.argument("events_field", type=str)
@click.option(
    "--events-category",
    type=str,
    help="Field with event category (categories are modeled as separate regressors)",
)
@click.option(
    "--sort-field",
    type=str,
    help="Colon-separated list of fields to sort by for output order",
)
@click.option("--high-pass", help="Highpass filter in Hz", type=float, default=0)
@click.option("--smooth", help="Smoothing kernel FWHM", type=float)
@click.option(
    "--confound-measures", help="Colon-separated list of confound measures to include"
)
@click.option(
    "--censor-motion/--no-censor-motion",
    default=False,
    help="Censor times around high-motion frames (default is no censoring)",
)
@click.option(
    "--censor-motion-range",
    type=int,
    nargs=2,
    default=(-1, 2),
    help="Range of frames to exclude around high motion (default is (-1, 2))",
)
def betaseries_bids(
    data_dir,
    fmriprep_dir,
    out_dir,
    subject,
    task,
    run,
    space,
    mask_name,
    mask_file,
    events_field,
    events_category,
    sort_field,
    high_pass,
    smooth,
    confound_measures,
    censor_motion,
    censor_motion_range,
):
    """
    Create a betaseries image estimating stimulus activity for a functional run.

    This script is designed to use BIDS-formatted data and data preprocessed
    using fMRIPrep.

    \b
    Required inputs
    ---------------
    data_dir
        Path to a BIDS-compliant dataset with task events.
    fmriprep_dir
        Path to fMRIPrep output.
    out_dir
        Path to main output directory.
    subject
        Subject ID.
    task
        Task code.
    run
        Run number.
    space
        Space in which the data should be analyzed (must have been a space output
        by fMRIPrep).
    mask_name
        Name of the mask to use when selecting voxels to analyze.
    mask_file
        Path to a binary mask indicating voxels to analyze.
    events_field
        Column of the events file to use to indicate individual EVs (explanatory
        variables) to model.
    """
    data_dir = Path(data_dir)
    fmriprep_dir = Path(fmriprep_dir)
    if sort_field is None:
        sort_field = events_field
    else:
        sort_field = sort_field.split(":")
    if confound_measures is not None:
        confound_measures = confound_measures.split(":")

    # task events, functional data, and confounds
    events_path = (
        data_dir
        / f"sub-{subject}"
        / "func"
        / f"sub-{subject}_task-{task}_run-{run}_events.tsv"
    )
    func_path = (
        fmriprep_dir
        / f"sub-{subject}"
        / "func"
        / f"sub-{subject}_task-{task}_run-{run}_space-{space}_desc-preproc_bold.nii.gz"
    )
    confound_path = (
        fmriprep_dir
        / f"sub-{subject}"
        / "func"
        / f"sub-{subject}_task-{task}_run-{run}_desc-confounds_timeseries.tsv"
    )

    # load functional timeseries information
    func_json = func_path.with_suffix("").with_suffix(".json")
    with open(func_json, "r") as f:
        func_param = json.load(f)
    tr = func_param["RepetitionTime"]
    time_offset = func_param["StartTime"]

    # create nuisance regressor matrix
    confounds = pd.read_table(confound_path)
    nuisance, nuisance_names = create_confound_matrix(
        confounds, confound_measures, censor_motion, censor_motion_range
    )

    # run betaseries estimation and save results
    run_betaseries(
        events_path,
        events_field,
        tr,
        time_offset,
        func_path,
        mask_name,
        mask_file,
        out_dir,
        subject,
        task,
        run,
        space,
        events_category=events_category,
        nuisance=nuisance,
        nuisance_names=nuisance_names,
        high_pass=high_pass,
        smooth_fwhm=smooth,
        sort_field=sort_field,
    )
